import numpy as np

from PyQt5.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QMainWindow,
    QScrollArea,
    QVBoxLayout,
    QWidget,
)

from ..widgets import LabelAndIntNumber, VStack
from .coincidence_plots import PlotCoincidenceTraces


class CoincidencePlot(QHBoxLayout):
    def __init__(self, main: QMainWindow):
        super().__init__()

        self.main = main

        # units in FPGA cycles
        self.maximum = 200
        self.channels = [0, 1, 2]
        self.N = len(self.channels)
        self.channel_widgets = []
        self.multiplicity = [
            0b0000_0000_0000_0000_0111,
            0b0000_0000_0000_0000_0011,
            0b0000_0000_0000_0000_0101,
        ]

        self.cursor_select = False

        self.plot_area = PlotCoincidenceTraces(
            main=self.main, N=self.N, maximum=self.maximum
        )

        self.plot_area.fig.canvas.mpl_connect("button_press_event", self.onclick)
        self.plot_area.fig.canvas.mpl_connect("key_press_event", self.onkey)

        self.scroll = QScrollArea()
        self.vvbox = QWidget()
        self.vbox = QVBoxLayout()
        self.update_channels()
        self.vvbox.setLayout(self.vbox)
        self.scroll.setWidget(self.vvbox)

        self.addWidget(self.plot_area)
        self.addWidget(self.scroll)

        self.draw()

    def update_channels(self):

        while self.vbox.count():
            item = self.vbox.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.vstack.delete()
                widget.vstack.deleteLater()
                widget.deleteLater()

        self.N = len(self.channels)
        self.plot_area.N = self.N
        self.plot_area.calc_positions()

        self.position = [0] * self.N

        self.trigger_delay = [10] * self.N
        self.trigger_stretch = [10] * self.N

        self.coincidence_position = [0] * self.N

        self.internal_delay = 10
        self.validation_delay = [10] * self.N
        self.validation_stretch = [10] * self.N

        self.FT = np.zeros((self.N, self.maximum))
        self.window = np.zeros((self.N, self.maximum))
        self.coincidence = np.ones((self.N, self.maximum))
        self.validation = np.zeros((self.N, self.maximum))
        self.delayed_FT = np.zeros((self.N, self.maximum))

        self.channel_widgets = []
        for c in self.channels:
            inputs = []
            inputs.append(LabelAndIntNumber("Trigger Delay"))
            inputs.append(LabelAndIntNumber("Trigger Stretch"))
            inputs.append(LabelAndIntNumber("Validation Delay"))
            inputs.append(LabelAndIntNumber("Validation Stretch"))
            inputs.append(LabelAndIntNumber("Multiplicity"))

            tmp = VStack(inputs, callback=self.update)
            groupbox = QGroupBox(f"Channel {c}")
            groupbox.vstack = tmp
            groupbox.setLayout(tmp)

            self.channel_widgets.append(groupbox)

        for i in self.channel_widgets:
            self.vbox.addWidget(i)

        self.draw()

    def onclick(self, event):
        y = event.ydata
        if y is None:
            return

        x = int(event.xdata)

        i = self.plot_area.get_y_range(y)
        if i is not None:
            self.position[i] = x

        self.draw()

    def onkey(self, event):
        y = event.ydata
        if y is None:
            return

        x = int(event.xdata)

        i = self.plot_area.get_y_range(y)
        if i is not None:
            if event.key == "left":
                self.position[i] -= 1
            if event.key == "right":
                self.position[i] += 1
            self.position[i] = min(max(self.position[i], 0), self.maximum)

        self.draw()

    def update(self):
        for i, box in enumerate(self.channel_widgets):
            values = box.vstack.value
            self.trigger_delay[i] = values[f"Trigger Delay"]
            self.trigger_stretch[i] = values[f"Trigger Stretch"]
            self.validation_delay[i] = values[f"Validation Delay"]
            self.validation_stretch[i] = values[f"Validation Stretch"]

        self.draw()

    def draw(self):
        self.plot_area.clear()

        self.FT *= 0
        self.window *= 0
        self.coincidence *= 0
        self.delayed_FT *= 0
        self.validation *= 0

        for i in range(self.N):
            self.FT[i, self.position[i]] = 1
            left = min(self.position[i] + self.trigger_delay[i], self.maximum)
            right = min(
                self.position[i] + self.trigger_delay[i] + self.trigger_stretch[i],
                self.maximum,
            )
            self.window[i, left:right] = 1
            # todo use mask
        for i in range(self.N):
            multiplicity = self.multiplicity[i]
            multiplicity = f"{multiplicity:016b}"
            # list of channels
            multiplicity = [i for i, c in enumerate(reversed(multiplicity)) if int(c)]
            for j, c in enumerate(self.channels):
                if c in multiplicity:
                    self.coincidence[i] += self.window[j]
            self.coincidence[i] = self.coincidence[i] >= 2
        self.coincidence_position = [trace.argmax() for trace in self.coincidence]
        for i in range(self.N):
            if self.coincidence_position[i] == 0:
                self.validation[i] *= 0
            else:
                left = min(
                    self.coincidence_position[i] + self.internal_delay, self.maximum
                )
                right = min(
                    self.coincidence_position[i]
                    + self.internal_delay
                    + self.validation_stretch[i],
                    self.maximum,
                )

                self.validation[i, left:right] = 1
            self.delayed_FT[i, self.position[i] + self.validation_delay[i]] = 1

        # print the fast trigger and the delayed and stretch version
        self.plot_area.plot_titles()

        for i, c in enumerate(self.channels):
            self.plot_area.plot_fast_trigger_and_coinc_window(
                self.FT[i], self.window[i], i, c
            )

            # print the coincidence trigger
        for i, c in enumerate(self.channels):
            self.plot_area.plot_coincidence(self.coincidence[i], i, c)

        # print the validation signals
        for i, c in enumerate(self.channels):
            self.plot_area.plot_validation(
                self.coincidence[i],
                self.validation[i],
                self.FT[i],
                self.delayed_FT[i],
                i,
                c,
            )

        self.plot_area.canvas.draw()



"""Functions to load settings from a config file."""

import configparser
import os
from pathlib import Path
import sys
from typing import List, Union

import appdirs


dirs = appdirs.AppDirs("PIXIE16")
inifile = Path(dirs.user_config_dir) / "config.ini"

config = configparser.ConfigParser()
config.read(inifile)


def config_get_parameters(
    section: str, name: str, verbose: bool = True
) -> Union[None, Path]:
    """Get the setting out of the config dir.

    Provide better error message if init file does not exist.

    """
    try:
        path_str = config.get(section, name).replace('"', "")
        path_str = path_str.replace("'", "")
        path = Path(path_str)
    except (configparser.NoOptionError, configparser.NoSectionError):
        path = None
        if verbose:
            print()
            print(f"No {name} found in {section}, please add it to {inifile}")
            print("   The file should contain something like:")
            print(f"       [{section}]")
            print(f"       {name} = <setting for {name}>")
            print()
            print("The file should contain the following sections and keys:")
            print("   [Libraries]")
            print("   sdk = <path to Pixie16Api.dll>")
            print("   [Data]")
            print("   datadir = <path where the data files should live>")
            print("   [Firmware.default]")
            print("   ComFPGAConfigFile = <path to syspixie16 firmware>")
            print("   SPFPGAConfigFile = <path to fippixie16 firmware>")
            print("   DSPCodeFile = <path to Pixie16DSP*.ldr>")
            print("   DSPVarFile = <path to Pixie16DSP*.var>")
            print("   SettingFile = <path to default setting file>")
            print()
    return path


def load_config(section_name: str) -> List[Path]:
    """Load firmware, etc setting that are defined in a section of the config file.

    Sections are defined with 'Firmware.<sectionname>'.
    """
    section = f"Firmware.{section_name}"
    if section not in config.sections():
        print("Error: cannot find section {section} in the config file {inifile}")
        sys.exit(1)
    # load new path to firmware
    firmware_com = config_get_parameters(section, "ComFPGAConfigFile")
    firmware_sp = config_get_parameters(section, "SPFPGAConfigFile")
    firmware_dsp_code = config_get_parameters(section, "DSPCodeFile")
    firmware_dsp_var = config_get_parameters(section, "DSPVarFile")
    setting_file = config_get_parameters(section, "SettingFile")

    return [
        firmware_com,
        firmware_sp,
        firmware_dsp_code,
        firmware_dsp_var,
        setting_file,
    ]


def list_firmware() -> None:
    """List all firmwars defined in the config file."""
    print(f"The config file used is: {inifile}")
    print("The following firmware definitions exists")
    names = []
    for section in config.sections():
        if not section.startswith("Firmware."):
            continue
        print(f"{section}")
        names.append(section[9:])
        for key in config[section].keys():
            print(f"   {key} = {config[section][key]}")
    print(
        f'Use only the name after the "." for the name of'
        f' the firmware: {", ".join(names)}'
    )


def get_sdk():
    sdk = config_get_parameters("Libraries", "sdk")

    # make sure the library directory in the path, so that we can find
    # dependencies, otherwise we get a "[WinError 126] The specified
    # module could not be found" error
    if sdk:
        os.environ["PATH"] = str(sdk.parent) + ";" + os.environ["PATH"]

    return str(sdk)



def read_mca_mode_data(filename):
    """Read MCA data files: 32k 32bit words for 16 channels"""

    results = {}
    with open(filename, "rb") as f:
        for channel in range(16):
            spectrum = []
            for i in range(32 * 1024):
                spectrum.append(int.from_bytes(f.read(4), byteorder="little"))
            spectrum = np.array(spectrum)
            results[channel] = spectrum
    return results



#!/usr/bin/env python3
"""
Usage:
  convert_settings.py FILENAME [MODULE_LIST ...] [options]
  convert_settings.py -h | --help

Arguments:
  FILENAME        Name of .set settings file to be converted
  MODULE_LIST     List of modules of interest

Options:
  -h --help        Show this screen.
  --metadata=FILE  Provide a .json file with metadata

Example:
  python convert_settings.py my_settings.set 0 2 3

The program converts a .set settings file to a .json settings file.
"""

import docopt
from pixie16.read.settings.old_settings_class import convert_old_settings

docopt_args = docopt.docopt(__doc__)

# Parse command line
filename = docopt_args["FILENAME"]
mod_list = docopt_args["MODULE_LIST"]
mod_list = list(map(int, mod_list))
metadata_file = docopt_args["--metadata"]

set_filename, stat_filename = convert_old_settings(filename, mod_list, metadata_filename=metadata_file)



"""pixie16-coincidence  helps setting up coincidence modes

Usage:
  pixie16-coincidence [<setting_filename>]

"""

from docopt import docopt


def main():
    commands = docopt(__doc__)

    from pathlib import Path
    import sys

    from pixie16.ui import CoincidenceApp, QApplication

    filename = commands["<setting_filename>"]
    filename = Path(filename) if filename else None
    if filename and not filename.exists():
        print(f"[Error] Cannot find file {filename}")
        sys.exit(1)

    app = QApplication([])
    ex = CoincidenceApp(app)
    if filename:
        ex.open_file(filename)
    sys.exit(app.exec_())



"""A collection of tasks that can be used in a pipeline.

These tasks can also easily by sub-classed to create custom ones or
used as an example on how to write your own.

"""
from collections import defaultdict, namedtuple
from datetime import datetime
from pathlib import Path
import time
from typing import List

import numpy as np

from .pipeline import Task
from . import control
from . import read


class DummyData(Task):
    """A class that will send dummy data into a pipeline.

    This can be used as a first task in a pipeline during testing and
    when the pixie16 is not online/available.
    """

    def __init__(self, runtime, filename=None):
        super().__init__()
        self.runtime = runtime
        self.filename = filename

        self.name = "Data generator"
        if filename is None:
            file = Path(__file__).parent.parent / "tests/pixie16-data-01.bin"
        else:
            file = Path(filename)
        self.data = np.fromfile(file, dtype=np.uint8)
        self.length = len(self.data)
        self.chunk_size = 20_000
        self.pos = 0
        if self.chunk_size > self.length:
            print("[ERROR] chunk_size too big")
        self.modules = [2]
        self.mycounter = 0

    def do_work(self, value):
        if time.time() - self.start_time > self.runtime:
            self.done = True
        time.sleep(0.2)
        self.mycounter += 1
        if self.pos + self.chunk_size < self.length:
            out = self.data[self.pos : self.pos + self.chunk_size]
        else:
            out = self.data[self.pos :]
            self.pos = 0
        self.pos += self.chunk_size
        ret = [out for m in self.modules]
        return ret


class TakeData(Task):
    """Task to aquire data, each binary blob from the FPGA will be put in the queue.

    Note: this does not work as is. The Task must also call InitSys
    and BootModule to be able to talk to the pixie16.
    """

    def __init__(self):
        super().__init__()
        print("save settings")
        self.name = "Take Data"

        control.start_list_mode_run()

    def do_work(self, value):
        return control.read_list_mode_fifo(threshold=64 * 1024)

    def cleanup(self):
        print("save settings")


class SetDefaults(Task):
    """Set defaults for parameters in the pixie."""

    def __init__(self, modules: List):
        super().__init__()
        self.name = "Set defaults"
        self.modules = modules

    def do_work(self, value):
        self.done = True
        control.init_and_boot(modules=self.modules)
        for m, _ in enumerate(self.modules):
            control.set_defaults_for_module(m)


class GatherData(Task):
    """Task to create larger data buckets out of the data directly from the FPGA.

    It has two different buckets: one for sending data to the next queue
    and one for saving data to disk.
    """

    def __init__(self, maxsize=50e6, save_size=None, path=None):
        super().__init__()
        self.data_queue = defaultdict(list)
        self.data_disk = defaultdict(list)
        self.maxsize = maxsize  # in bytes
        self.save_size = save_size or maxsize
        self.save_binary = save_size is not None
        self.name = "Gather Data"
        self.mycounter = 0
        self.file_counter = 0
        self.path = path or Path(".")

    def save_data(self, out):
        timestamp = datetime.today().strftime("%Y-%m-%d-%H-%M-%S")
        for k, v in out.items():
            file = self.path / f"data-mod{k}-{timestamp}-{self.file_counter:05d}.bin"
            v.tofile(str(file))
        self.file_counter += 1

    def get_size(self, data):
        out = []
        for data_list in data.values():
            size = 0
            for element in data_list:
                size += element.nbytes
            out.append(size)
        if out:
            return max(out)
        return 0

    def do_work(self, value):
        self.mycounter += 1

        for i, data in enumerate(value):
            self.data_queue[i].append(data)
            self.data_disk[i].append(data)

        # handle data for saving
        data_size = self.get_size(self.data_disk)
        if data_size:
            self.send_status({"data disk size": data_size})
            if data_size > self.save_size:
                out = {k: np.concatenate(v) for k, v in self.data_disk.items() if v}
                if self.save_binary:
                    self.save_data(out)
                self.data_disk = defaultdict(list)

        # handle data for queue
        data_size = self.get_size(self.data_queue)
        self.send_status({"data queue size": data_size})

        if not data_size:
            return

        if data_size > self.maxsize:
            out = {k: np.concatenate(v) for k, v in self.data_queue.items() if v}
            self.data_queue = defaultdict(list)
            return out

    def cleanup(self):
        out_queue = {k: np.concatenate(v) for k, v in self.data_queue.items() if v}
        out_disk = {k: np.concatenate(v) for k, v in self.data_disk.items() if v}
        if out_queue:
            if self.output_queue:
                self.output_queue.put(out_queue)
        if out_disk:
            if self.save_binary:
                self.save_data(out_disk)


class ConvertToEvents(Task):
    """Task to convert data stream to events."""

    def __init__(self):
        super().__init__()
        self.list_mode_readers = {}
        self.name = "Convert to events"

    def do_work(self, value_dict):
        for i, v in value_dict.items():
            if i not in self.list_mode_readers:
                self.list_mode_readers[i] = read.ListModeDataReader()
            self.list_mode_readers[i].put(v.tobytes())
        return {mod: reader.pop_all() for mod, reader in self.list_mode_readers.items()}


class PickSingleModule(Task):
    """Task to pick events from a single module.

    Takes output from, e.g., ConvertToEvents and outputs only the data
    for a single module.
    """

    def __init__(self, module=0):
        super().__init__()
        self.module = module
        self.name = f"Reduce to module {module}"

    def do_work(self, value):
        return value[self.module]


class SortEvents(Task):
    """Task to sort events by timestamp."""

    def __init__(self, maxsize=10_000, number_to_sort=8_000):
        super().__init__()
        self.data = []
        assert (
            number_to_sort < maxsize
        ), "The number_to_sort needs to be smaller than maxisze"
        self.maxsize = maxsize
        self.N_to_sort = number_to_sort
        self.name = "Sort events"
        self.nr_sorted = 0

    def do_work(self, events_lst):
        # Convert ns timestamp into an approximate UNIX time
        # This needs to be done because the ns timestamps are reset after each MCA run
        current_time = time.time()
        new_events_lst = []
        for event in events_lst:
            new_event = event._replace(
                chunk_timestamp=current_time
                + 1e-9 * (event.timestamp - events_lst[0].timestamp)
            )
            new_events_lst.append(new_event)

        # Sort events according to UNIX timestamp
        out = []
        self.data.extend(new_events_lst)
        while len(self.data) > self.maxsize:
            self.data.sort(key=lambda x: x.chunk_timestamp)
            out = self.data[: self.N_to_sort]
            self.data = self.data[self.N_to_sort :]
            if out and self.output_queue:
                self.output_queue.put(out)
                self.nr_sorted += 1
        self.send_status({"sorted blocks": self.nr_sorted})

    def cleanup(self):
        self.data.sort(key=lambda x: x.chunk_timestamp)
        if self.output_queue:
            self.output_queue.put(self.data)
        self.data = []


class GatherEvents(Task):
    """Gather Events into larger chunks."""

    def __init__(self, size=1_000_000):
        super().__init__()
        self.data = []
        self.size = size
        self.nr = 0
        self.name = f"Gather events (size={size})"

    def do_work(self, value):
        self.data.extend(value)
        if len(self.data) > self.size:
            out = self.data
            self.data = []
            self.nr += 1
            return out

        self.send_status({"gathered events": self.nr, "gathered queue": len(self.data)})

    def cleanup(self):
        if self.output_queue and self.data:
            self.output_queue.put(self.data)
        self.data = []
        self.nr += 1
        self.send_status({"gathered events": self.nr, "gathered queue": len(self.data)})


class LoadFiles(Task):
    """Load events from a list of files,"""

    def __init__(self, file_list, batch_size=1_000):
        super().__init__()
        self.files = file_list
        self.byte_stream = read.FilesIO(self.files)
        self.buffer_size = 1_000_000
        self.reader = read.ListModeDataReader()
        self.number_of_events_to_read = batch_size
        self.nr_of_events = 0
        self.name = "Loading binary data from files"

    def do_work(self, value):
        try:
            events = []
            for _ in range(self.number_of_events_to_read):
                event = None
                try:
                    event = self.reader.pop()
                except (read.LeftoverBytesError, read.EmptyError):
                    if not self.byte_stream.is_empty():
                        self.reader.put(self.byte_stream.pop(self.buffer_size))
                        continue
                    self.done = True
                    break
                events.append(event)
            if events and self.output_queue:
                self.output_queue.put(events)
                self.nr_of_events += len(events)
        except StopIteration:
            self.done = True

        self.send_status(
            {
                "read events": f"{self.nr_of_events:.2e}",
                "runtime": self.byte_stream.current_file_index,
            }
        )



from pathlib import Path

import pytest

import pixie16


@pytest.fixture(scope="module")
def settings_file():
    return Path(__file__).parent / "pixie16-settings.set"


class TestReadSettingsData:
    def test_read_settings(self, settings_file):
        s = pixie16.read.load_settings(settings_file)

        livetime = s.get("LiveTime", channel=3, as_pint=True)
        cfd_delay = s.get("CFDDelay", channel=9, as_pint=True)
        trace_delay = s.get("TraceDelay", channel=3, as_pint=True)
        trace_length = s.get("TraceLength", channel=0, as_pint=True)
        ext_delay_length = s.get("ExternDelayLen", channel=0, as_pint=True)

        assert round(livetime.to("seconds").magnitude, 2) == 599.82
        assert s["LiveTime"][9] == 59981879462
        assert s.FastGap[3] == 10
        assert s.Log2Ebin[3] == 4294967295
        assert s["CFDThresh"][0] == 120
        assert cfd_delay.to("seconds").magnitude == 8e-8
        assert s["BLcut"][0] == 222
        assert s["BaselinePercent"][0] == 10
        assert trace_delay.to("seconds").magnitude == 2e-7
        assert s["TraceLength"][0] == 5000
        assert trace_length.to("seconds").magnitude == 1e-5
        assert s["ExternDelayLen"][0] == 100
        assert ext_delay_length.to("seconds").magnitude == 1e-6



"""This file defines the default units for each element in the setting class."""


SETTINGS_UNITS_TBL = {
    "OffsetVoltage": "V",
    "TraceDelay": "FPGA_cycles",
    "FastPeaks": "counts",
    "RealTime": "FPGA_cycles",
    "LiveTime": "FPGA_cycles",
    "RunTime": "FPGA_cycles",
    "ExternDelayLen": "FPGA_cycles",
    "ModNum": "unknown",
    "ModCSRA": "bits",
    "ModCSRB": "bits",
    "ModFormat": "unknown",
    "RunTask": "unknown",
    "ControlTask": "unknown",
    "MaxEvents": "unknown",
    "CoincPattern": "unknown",
    "CoincWait": "unknown",
    "SynchWait": "bits",
    "InSynch": "bits",
    "Resume": "unknown",
    "SlowFilterRange": "unknown",
    "FastFilterRange": "unknown",
    "ChanNum": "unknown",
    "HostIO": "unknown",
    "UserIn": "unknown",
    "FastTrigBackplaneEna": "unknown",
    "CrateID": "unknown",
    "SlotID": "unknown",
    "ModID": "unknown",
    "TrigConfig": "bits",
    "U00": "unknown",
    "HostRunTimePreset": "unknown",
    "PowerUpInitDone": "flag",
    "ChanCSRa": "bits",
    "ChanCSRb": "bits",
    "GainDAC": "unknown",
    "OffsetDAC": "unknown",
    "DigGain": "unknown",
    "SlowLength": "slow_filter_cycles",
    "SlowGap": "slow_filter_cycles",
    "FastLength": "fast_filter_cycles",
    "FastGap": "fast_filter_cycles",
    "PeakSample": "slow_filter_cycles",
    "PeakSep": "FPGA_cycles",
    "CFDThresh": "CDF_filter_amplitude",
    "FastThresh": "fast_filter_amplitude",
    "ThreshWidth": "unknown",
    "PAFlength": "FPGA_cycles",
    "TriggerDelay": "FPGA_cycles",
    "ResetDelay": "unknown",
    "ChanTrigStretch": "FPGA_cycles",
    "TraceLength": "ADC_cycles",
    "Xwait": "unknown",
    "TrigOutLen": "unknown",
    "EnergyLow": "slow_filter_amplitude",
    "Log2Ebin": "unknown",
    "MultiplicityMaskL": "bits",
    "PSAoffset": "unknown",
    "PSAlength": "unknown",
    "Integrator": "unknown",
    "BLcut": "ADC_amplitude",
    "BaselinePercent": "unknown",
    "FtrigoutDelay": "unknown",
    "Log2Bweight": "unknown",
    "PreampTau": "us",
    "Xavg": "unknown",
    "MultiplicityMaskH": "bits",
    "FastTrigBackLen": "FPGA_cycles",
    "CFDDelay": "FPGA_cycles",
    "CFDScale": "unknown",
    "ExtTrigStretch": "FPGA_cycles",
    "VetoStretch": "FPGA_cycles",
    "ExternDelayLen": "FPGA_cycles",
    "QDCLen0": "unknown",  # FPGA or ADC cycles
    "QDCLen1": "unknown",
    "QDCLen2": "unknown",
    "QDCLen3": "unknown",
    "QDCLen4": "unknown",
    "QDCLen5": "unknown",
    "QDCLen6": "unknown",
    "QDCLen7": "unknown",
    "GSLTtime": "unknown",
    "NumEventsA": "unknown",
    "NumEventsB": "unknown",
    "DSPerror": "unknown",
    "SynchDone": "unknown",
    "BufHeadLen": "unknown",
    "EventHeadLen": "unknown",
    "ChanHeadLen": "unknown",
    "UserOut": "unknown",
    "AOutBuffer": "unknown",
    "LOutBuffer": "unknown",
    "AECorr": "unknown",
    "LECorr": "unknown",
    "HardwareID": "unknown",
    "HardVariant": "unknown",
    "FIFOLength": "unknown",
    "FippiID": "unknown",
    "FippiVariant": "unknown",
    "DSPrelease": "unknown",
    "DSPbuild": "unknown",
    "DSPVariant": "unknown",
    "U20": "unknown",
    "OverflowA": "unknown",
    "OverflowB": "unknown",
    "InSpecA": "unknown",
    "InSpecB": "unknown",
    "UnderflowA": "unknown",
    "UnderflowB": "unknown",
    "ChanEventsA": "unknown",
    "ChanEventsB": "unknown",
    "AutoTau": "unknown",
    "U30": "unknown",
}



import numpy as np
from matplotlib import cm
from rich import print

from PyQt5.QtWidgets import (
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QWidget,
)


from ..plot import Plot
from ..widgets import LabelAndIntNumber, VStack
from ...analyze import (
    convert_trace_to_FPGA,
    calculate_fast_filter,
    get_fast_trigger_index,
)


class FastTrigger(QWidget):
    """Create all kind of traces that are calculated on the FPGA."""

    def __init__(self, main):
        super().__init__()
        self.main = main

        self.buttons = QHBoxLayout()
        self.buttons_widgets = []
        self.plot_names = [
            "traces",
            "fast trigger",
            "CFD",
            "energy",
            "energy linearity",
        ]
        self.buttons.addWidget(QLabel("Show/Hide:"))
        for i in self.plot_names:
            tmp = QPushButton(i)
            tmp.setCheckable(True)
            tmp.toggle()
            tmp.clicked.connect(
                lambda state, name=i, widget=tmp: self.show_hide(name, widget)
            )
            self.buttons_widgets.append(tmp)
            self.buttons.addWidget(tmp)

        self.grid = QGridLayout()

        self.trace_plot = Plot(main=main)
        self.fast_plot = Plot(main=main)
        self.cfd_plot = Plot(main=main)
        self.energy_plot = Plot(main=main)
        self.energy_linearity_plot = Plot(main=main)

        self.N = 10
        self.raw_data = np.zeros(self.N)[:, np.newaxis]
        self.data_FPGA = np.zeros(self.N)[:, np.newaxis]
        self.fast_trigger = np.zeros(self.N)[:, np.newaxis]
        # get N colors for plotting
        self.colors = cm.get_cmap("jet")(np.linspace(0, 1, self.N))
        self.triggers = np.zeros(self.N)
        self.baseline_index = 1
        self.energy_values = np.zeros(self.N)

        self.channel_label = QLabel("Channel")
        self.channels = QComboBox()
        self.channels.currentIndexChanged.connect(self.pick_channel)

        self.ft = VStack(
            [
                LabelAndIntNumber("G"),
                LabelAndIntNumber("L"),
                LabelAndIntNumber("Threshold", value=100),
            ],
            callback=self.update_fasttrigger,
        )

        self.cfd = VStack([LabelAndIntNumber("Threshold", value=100)])

        self.energy = VStack(
            [
                LabelAndIntNumber("Range", name="R"),
                LabelAndIntNumber("G (energy steps)", name="G"),
                LabelAndIntNumber("L (energy steps)", name="L"),
                LabelAndIntNumber(
                    "Peaksample (energy steps)", value=10, name="peaksample"
                ),
                LabelAndIntNumber("decay (ns)", value=28, name="decay"),
            ],
            callback=self.update_energies,
        )

        self.energy_linearity = VStack(
            [LabelAndIntNumber("Integration Length", name="L")],
            callback=self.update_energy_linearity,
        )

        self.grid.addLayout(self.buttons, 0, 0, 1, 2)

        self.grid.addWidget(self.trace_plot, 1, 0)
        self.grid.addWidget(self.fast_plot, 2, 0)
        self.grid.addWidget(self.cfd_plot, 3, 0)
        self.grid.addWidget(self.energy_plot, 4, 0)
        self.grid.addWidget(self.energy_linearity_plot, 5, 0)

        self.grid.addWidget(self.channels, 1, 1)
        self.grid.addLayout(self.ft, 2, 1)
        self.grid.addLayout(self.cfd, 3, 1)
        self.grid.addLayout(self.energy, 4, 1)
        self.grid.addLayout(self.energy_linearity, 5, 1)

        # todo: add channel select and index select at bottom
        # add L, G, threshold  for CFD and FAST trigger and energy

        # add other tab to show histogram across all events of peaks or CFD

        self.setLayout(self.grid)

    def show_hide(self, name, widget):
        for i, plot_name in enumerate(self.plot_names):
            if name == plot_name:
                for c in [0, 1]:
                    tmp = self.grid.itemAtPosition(1 + i, c).widget()
                    if tmp is None:
                        print(f' no items at {i+1} {c} for {name}"')
                        continue
                    tmp.setHidden(not tmp.isHidden())

    def add_channels(self):
        for c in self.main.channels:
            self.channels.addItem(str(c))

    def update_traces(self, channel):
        """Show the traces in FPGA cycles.

        We pick 10 traces for 10 different energies in the channels,
        plot them and store them, so they can be used for the other
        plots.

        """

        events = [
            d for d in self.main.data if d.channel == channel and len(d.trace) > 0
        ]
        if not len(events):
            print(f"[yellow]INFO[/] no events for channel {channel} that have traces")
            return
        events.sort(key=lambda x: x.energy)

        # pick N events equally spaced in energy sorted event list
        N = np.linspace(0, 1, self.N) * (len(events) - 1)
        N = sorted({int(n) for n in N})

        self.trace_plot.axes.clear()
        self.raw_data = np.array([events[n].trace for n in N])
        self.data_FPGA = convert_trace_to_FPGA(self.raw_data)
        for i, ft in enumerate(self.data_FPGA):
            self.trace_plot.axes.plot(ft, color=self.colors[i], drawstyle="steps-mid")
        self.trace_plot.axes.set_title("traces")
        self.trace_plot.axes.set_xlabel("FPGA cycles")
        self.trace_plot.canvas.draw()

    def update_fasttrigger(self):
        """Calculate and show the Fast Trigger signal and the threshold.

        We also try to give a good estimmate for the trigger
        threshold, but this is only based on the 10 traces.  For this
        we calcualte the standard derivation and maximum value from
        all traces up to 20 ns before the first calculated trigger.

        Those values will also be used to estimate the baseline for the energy calculation.
        The triggers are also stored, so that they can be re-used for CFD and energy calculation.

        """

        L = self.ft.value["L"]
        G = self.ft.value["G"]
        T = self.ft.value["Threshold"]

        self.triggers *= 0

        self.fast_plot.axes.clear()
        if self.data_FPGA.max() == 0:
            return

        length = len(self.data_FPGA[0])

        self.fast_triggers = calculate_fast_filter(self.raw_data, L=L, G=G)

        self.triggers = get_fast_trigger_index(self.fast_triggers, T)

        # plots traces
        for i, trace in enumerate(self.fast_triggers):
            self.fast_plot.axes.plot(trace, color=self.colors[i], drawstyle="steps-mid")

        # plot triggers
        for i, positions in enumerate(self.triggers):
            for p in positions:
                self.fast_plot.axes.axvline(p, color=self.colors[i])

        # plot threshold
        self.fast_plot.axes.axhline(T, color="black")

        # calc noise level
        try:
            first_trigger_offset = int(min(t[0] for t in self.triggers) - 10)
        except (ValueError, IndexError):
            print(
                "[orange3]WARNING[/] No trigger found, using the first 20% of the trace to estimate baseline",
                flush=True,
            )
            first_trigger_offset = int(length * 0.2)
        self.baseline_index = first_trigger_offset

        if first_trigger_offset > 0:
            average_noise = [r[:first_trigger_offset].std() for r in self.fast_triggers]
            average_noise = np.asarray(average_noise).mean()

            max_noise = [r[:first_trigger_offset].max() for r in self.fast_triggers]
            max_noise = np.asarray(max_noise).max()

            self.fast_plot.axes.set_title(
                f"Fast Trigger: noise level $\\sigma$={average_noise:.1f} "
                f"max={max_noise:.1f} before x={first_trigger_offset}"
            )
        else:
            self.fast_plot.axes.set_title("Fast Trigger: cannot calculate noise level")
            self.baseline_index = 1

        self.fast_plot.axes.set_xlabel("FPGA cycles")
        self.fast_plot.canvas.draw()

        self.update_energies()

    def update_energies(self):
        """Calcuate energy distribution as described in Mauricio's thesis.

        The energy calculation here is based on 20 ns time steps so
        currently only SlowFilterRange=1 works.

        See page 33 of Mauricio's thesis for details.

        Note: the pixie16 multiplies all energies by a factor of 4

        """
        R = self.energy.value["R"]
        L = self.energy.value["L"]
        G = self.energy.value["G"]
        P = self.energy.value["peaksample"]
        q = self.energy.value["decay"]
        q = np.exp(-2 / q)

        self.energy_values *= 0

        self.energy_plot.axes.clear()

        self.slow_filter_range = 2 ** R
        # check how long the trace is, if it is too short, we try to
        # fix it by adding data points at `baseline` value to the
        # trace on the left and right
        length = len(self.data_FPGA[0]) // self.slow_filter_range
        pad_data = False

        if length - 2 * L - G + 1 <= 0:
            print(
                "[orange3]WARNING[/] L, G for energy filter are too large for the supplied trace length and slowfilter range. Padding data",
                flush=True,
            )
            pad_data = True
            pad_length = 2 * (L + G)
            length += 2 * pad_length
        elif length - 2 * L - G + 1 <= 20:
            print(
                "[orange3]WARNING[/] L, G for energy filter are large for the supplied trace length and slowfilter range. Padding data",
                flush=True,
            )
            pad_data = True
            pad_length = L + G
            length += 2 * pad_length

        result = np.zeros(length - 2 * L - G + 1)

        qL = q ** (5 * L * self.slow_filter_range)
        c2 = 1 - q
        c3 = (1 - q) / (1 - qL)
        c1 = c3 * qL

        # self.data is already in FPGA cycles
        baseline = [t[: self.baseline_index].mean() for t in self.data_FPGA]
        # convert to energy cycles
        baseline_pad = np.asarray(baseline).mean() * self.slow_filter_range
        baseline = baseline_pad * (L + G) * 4 * c2

        for i, trace in enumerate(self.data_FPGA):
            # need to convert to avoid overflows in the sums
            trace = trace.astype(np.int64)

            # for the energy calculation groups of 10 ns * 2**slowfilterrange
            # the traces are already in 10ns groups, so we now handle the slowfilterrange
            N = (trace.shape[0] // self.slow_filter_range) * self.slow_filter_range
            trace = trace[:N].reshape(
                trace.shape[0] // self.slow_filter_range, self.slow_filter_range
            )
            trace = trace.sum(axis=1)

            if pad_data:
                pad = np.ones(pad_length) * int(baseline_pad)
                trace = np.concatenate([pad, trace, pad])

            if len(self.triggers[i]):
                # calculate trigger position in new time units
                trigger_pos = (
                    int(self.triggers[i][0]) // self.slow_filter_range - 2 * L - G + P
                )
                if pad_data:
                    trigger_pos += pad_length

            for k in range(2 * L + G - 1, length):
                S1 = trace[k - 2 * L - G + 1 : k - L - G + 1].sum()
                S2 = trace[k - L - G + 1 : k - L + 1].sum()
                S3 = trace[k - L + 1 : k + 1].sum()

                result[k - 2 * L - G + 1] = (
                    4 * (-c1 * S1 + c2 * S2 + c3 * S3) - baseline
                )
                if len(self.triggers[i]):
                    if k - 2 * L - G + 1 == trigger_pos:
                        print(
                            f"[yellow]INFO[/] found peaksample position {i}: ",
                            S1,
                            S2,
                            S3,
                            baseline,
                        )
                        self.energy_values[i] = result[k - 2 * L - G + 1]

            self.energy_plot.axes.plot(
                result, color=self.colors[i], drawstyle="steps-mid"
            )
            if len(self.triggers[i]):
                if 0 < trigger_pos < len(result):
                    self.energy_plot.axes.plot(
                        trigger_pos, result[trigger_pos], "o", color=self.colors[i]
                    )
            self.energy_plot.axes.set_xlabel(f"{self.slow_filter_range*10} ns cycles")
            self.energy_plot.axes.set_ylabel("Energy")
            self.energy_plot.axes.set_title(f"Energy baseline {baseline:.1f}")
        self.energy_plot.canvas.draw()

        self.update_energy_linearity()

    def update_energy_linearity(self):
        L = self.energy_linearity.value["L"]

        baseline = [t[: self.baseline_index].mean() for t in self.data_FPGA]
        baseline = np.asarray(baseline).mean()

        energy = np.zeros(self.N)
        for i, trace in enumerate(self.data_FPGA):
            try:
                trigger_pos = int(self.triggers[i][0])
            except IndexError:
                trigger_pos = 0
            if trigger_pos:
                if trigger_pos + L < len(trace):
                    energy[i] = (
                        trace[trigger_pos : trigger_pos + L].sum() - baseline * L
                    )
        self.energy_linearity_plot.axes.clear()
        X = energy
        Y = self.energy_values
        fit = np.poly1d(np.polyfit(X, Y, 1))
        F = fit(X)
        mask = X > 0
        if mask.sum() == 0:
            self.energy_linearity_plot.canvas.draw()
            return
        X = X[mask]
        Y = Y[mask]
        F = F[mask]
        C = np.array(self.colors)[mask]
        r2 = ((F - Y.mean()) ** 2).sum() / ((Y - Y.mean()) ** 2).sum()
        if r2 < 0.96:
            self.energy_linearity_plot.axes.plot(X, F, color="black")
            self.energy_linearity_plot.axes.scatter(X, Y, c=C)
            self.energy_linearity_plot.axes.set_xlabel("Integrated trace")
            self.energy_linearity_plot.axes.set_ylabel("Calculated energy")
            self.energy_linearity_plot.axes.set_title(f"Fit r2={r2:.6f}")
        else:
            self.energy_linearity_plot.axes.scatter(X, (Y - F) / F, c=C)
            self.energy_linearity_plot.axes.set_xlabel("Integrated energy")
            self.energy_linearity_plot.axes.set_ylabel("Error to linear fit [%]")
            self.energy_linearity_plot.axes.set_title(f"Fit Error  r2={r2:.6f}")

        self.energy_linearity_plot.canvas.draw()

    def pick_channel(self, i):
        channel = self.main.channels[i]

        self.update_traces(channel)
        self.update_fasttrigger()  # also updated energies



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))
on_rtd = os.environ.get("READTHEDOCS") == "True"

# ctypes doesn't seem to work with autodoc_mock_imports...and we have
# to manually import it so that it doesn't try to import the dll
from unittest.mock import MagicMock


class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


MOCK_MODULES = [
    "ctypes",
]
sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)

# for all other libraries we can let autodoc know to mock them.
autodoc_mock_imports = [
    "cbitstruct",
    "fast_histogram",
    "numpy",
    "pandas",
    "pint",
    "matplotlib",
    "rich",
    "PyQt5",
]
# -- Project information -----------------------------------------------------

project = "pixie16"
copyright = "2019-2022, The Regents of the University of California through Lawrence Berkeley National Laboratory (subject to receipt of any required approvals from the U.S. Dept. of Energy). All rights reserved"
author = "Arun Persaud et al."

# The full version, including alpha/beta/rc tags
from pkg_resources import get_distribution, DistributionNotFound

try:
    release = get_distribution("pixie16").version
except DistributionNotFound:
    release = "unkown"

# -- General configuration ---------------------------------------------------

# seems to be needed for RTD
master_doc = "index"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "sphinx.ext.napoleon"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

from recommonmark.parser import CommonMarkParser


class CustomCommonMarkParser(CommonMarkParser):
    def visit_document(self, node):
        pass


def setup(app):
    app.add_source_parser(CustomCommonMarkParser)



from pathlib import Path
import sys
from typing import List

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QIntValidator
from PyQt5.QtWidgets import (
    QAction,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QMainWindow,
    QMessageBox,
    QVBoxLayout,
    QWidget,
)

import pixie16


class LabelAndIntNumber(QWidget):
    """A generic Int input widget in combination with a QLabel."""

    def __init__(self, label: str, value: int = 1, name=None) -> None:
        super().__init__()
        if name is None:
            self.name = label
        else:
            self.name = name

        self.box = QHBoxLayout()

        self.label = QLabel(label)
        self.number = QLineEdit()
        self.validator = QIntValidator(1, 500_000)
        self.number.setValidator(self.validator)

        self.box.addWidget(self.label)
        self.box.addWidget(self.number)

        self.setLayout(self.box)

        self.value = value

    @property
    def value(self) -> int:
        return int(self.number.text())

    @value.setter
    def value(self, value):
        self.number.setText(str(int(value)))


class VStack(QVBoxLayout):
    """Stack of LabelAndIntNumber.

    Helper class to make it easier to create a VBoxLayout of
    LabelAndIntNumber Widgets and to connect a callback.

    """

    def __init__(self, widgets: List[LabelAndIntNumber], callback=None):
        super().__init__()
        self.widgets = widgets

        for w in widgets:
            self.addWidget(w)
            if callback:
                w.number.editingFinished.connect(callback)

    @property
    def value(self):
        return {w.name: w.value for w in self.widgets}

    @value.setter
    def value(self, value_dict):
        for k, v in value_dict.items():
            for w in self.widgets:
                if w.name == k:
                    w.value = v
                    break

    def delete(self):
        for w in self.widgets:
            w.deleteLater()


class BaseApp(QMainWindow):
    """Base class for pypixie16 QT apps.

    Sets up the main window, the title, a menubar and an info_bar.
    """

    def __init__(self, name="", app=None):
        super().__init__()
        self.app = app
        self.name = name
        self.data_path = Path.home()

        if sys.platform == "win32":
            QIcon.setThemeSearchPaths([r"c:\Program Files\Git\usr\share\icons"])

        QIcon.setThemeName("Adwaita")

        # set up the window
        self.setWindowTitle(self.name)

        # file menu
        self.menu = self.menuBar()
        self.file_menu = self.menu.addMenu("File")
        self.help_menu = self.menu.addMenu("Help")

        quitButton = QAction(QIcon.fromTheme("application-exit"), "Exit", self)
        quitButton.setShortcut("Ctrl+Q")
        quitButton.setStatusTip("Quit")
        quitButton.triggered.connect(self.close)
        self.file_menu.addAction(quitButton)

        aboutButton = QAction(QIcon.fromTheme("help-about"), "About", self)
        aboutButton.setStatusTip("About")
        aboutButton.triggered.connect(self.about)
        self.help_menu.addAction(aboutButton)

        self.info_bar = QLabel()
        self.info_bar.setTextInteractionFlags(Qt.TextSelectableByMouse)

    def about(self):
        """Print some help that is useful for debugging"""
        pyversion = ".".join(str(i) for i in sys.version_info)
        QMessageBox.about(
            self,
            "About",
            f"{self.name}\n\n"
            + f"python-version: {pyversion}\n"
            + f"pixie16-version: {pixie16.__version__}",
        )



"""
Usage:
  change_setting.py set <name> <value> <module> <channel> <file> [<outfile>]
  change_setting.py read <name> <module> <file>
  change_setting.py load <file>
  change_setting.py list

"""

import docopt

commands = docopt.docopt(__doc__)
# print(commands)

name = commands["<name>"]
value = commands["<value>"]
module = commands["<module>"]
channel = commands["<channel>"]

# slots in the crate that have pixie16 cards
SLOTS = [2]
modules = [0]

import pixie16
import pixie16.control
import sys
import time
from pathlib import Path

settingfile = Path(commands["<file>"])
# r'E:\ROOTS\XIA-pulse-shape\2019-04-08-API-Cblock0013.set'
outfile = Path(commands["<outfile>"])

if commands["set"]:
    if not name in pixie16.variables.settings:
        print(f"name {name} not a valid settings name")
        sys.exit(1)

    value = int(value)
    if "-" in channel:
        start, stop = channel.split("-")
        channel = list(range(int(start), int(stop) + 1))
    else:
        channel = [int(channel)]
    module = int(module)

    if outfile:
        out = Path(outfile)
        if out.exists():
            print("Error: outfile exists already")
            sys.exit(2)

    pixie16.control.init_and_boot(
        offline_mode=False,
        section_name="default",
        modules=SLOTS,
    )
    settings = pixie16.control.read_settings_from_file(settingfile)
    for c in channel:
        pixie16.control.change_setting(settings, name, value, c, module)
    pixie16.control.write_settings(settings)
    if outfile:
        pixie16.control.save_dsp_parameters_as_json(outfile)
    pixie16.control.exit_system()

elif commands["read"]:
    if not name in pixie16.variables.settings:
        print(f"name {name} not a valid settings name")
        sys.exit(1)

    module = int(module)

    pixie16.control.init_and_boot(
        offline_mode=False,
        section_name="default",
        modules=SLOTS,
    )
    pixie16.control.load_dsp_parameters_from_file(settingfile)
    A = pixie16.control.read_settings_from_file(module)
    offset, nr = pixie16.variables.settings[name]
    print(f"Name: {name}")
    if nr > 1:
        print(f"Channel, value")
        for i in range(nr):
            print(f"    {i}  {A[offset+i]}")
    else:
        print(f"Module setting")
        print(f"      {A[offset]}")

    pixie16.control.exit_system()

elif commands["load"]:
    print(f"Loading settingfile {settingfile}")
    pixie16.control.init_and_boot(
        offline_mode=False,
        section_name="default",
        modules=SLOTS,
    )
    pixie16.control.load_dsp_parameters_from_file(settingfile)
    pixie16.control.exit_system()

elif commands["list"]:
    for n in sorted(pixie16.variables.settings):
        print(n)



from PyQt5.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QPushButton,
)


class ChannelSelect(QHBoxLayout):
    """Horizontal list of buttons that can be changed during runtime."""

    def __init__(self, main):
        super().__init__()
        self.channels = []
        self.main = main

    def clear_channels(self):
        while self.count():
            item = self.takeAt(0)
            widget = item.widget()
            if widget is not None:
                widget.deleteLater()

    def add_channels(self):
        self.addWidget(QLabel("Show/Hide:"))
        for channel in self.main.channels:
            tmp = QPushButton(f"Ch {channel}")
            tmp.setCheckable(True)
            tmp.toggle()
            tmp.clicked.connect(
                lambda state, channel=channel, widget=tmp: self.main.click_channel(
                    channel, widget
                )
            )

            self.channels.append(tmp)
            self.addWidget(tmp)
            # by default show all channels
            if channel not in self.main.selected_channels:
                self.main.selected_channels.append(channel)

        tmp = QPushButton("All")
        tmp.setCheckable(True)
        tmp.toggle()
        tmp.clicked.connect(
            lambda state, channel="all", widget=tmp: self.main.click_channel(
                "all", widget
            )
        )

        self.channels.append(tmp)
        self.addWidget(tmp)

    def update_button_status(self):
        for w, channel in zip(self.channels, self.main.channels):
            if not w.isChecked() and channel in self.main.selected_channels:
                self.main.dataview.clearSelection()
                w.toggle()
            if w.isChecked() and channel not in self.main.selected_channels:
                w.toggle()



from functools import wraps
import random
from typing import Union, List, Tuple, Optional, Sequence, Iterable
import warnings

from fast_histogram import histogram2d
import numpy as np
import matplotlib
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from matplotlib.backends.backend_pdf import PdfPages
from rich import print

from .analyze import calculate_CFD_using_FF
from .read import Event, Settings, Stats


DATA_POINTS_PER_FPGA_CYCLE = 5


def filter_events(
    events: Iterable[Event], N: Optional[int] = None, randomize: bool = False
) -> Iterable[Event]:
    """Pick N (random) events from a list of events"""

    N_events = len(events)

    if N is None:
        return events

    if N_events > N:
        if randomize:
            return random.sample(events, N)
        return events[:N]

    print(
        f"[orange3]WARNING[/] plotting: got less events than requested for plotting {N_events} < {N}"
    )
    return events


def create_title_page(
    title: Union[str, List[str]], pdf: Optional[PdfPages] = None
) -> None:
    """Helper function to create a title page

    The function will create a title centered in a plot.
    If pdf is given, it will save the figure inside the pdf (to be used with PdfPages)
    for multi page pdfs.

    Parameters
    ----------

    title
       The title, if a list of str, then each will be put on a single line
    pdf
       The pdf where the title page should be added to.

    """
    if isinstance(title, (list, tuple)):
        title = "\n".join(title)

    fig = plt.figure()
    plt.axis("off")
    plt.text(0.5, 0.5, title, ha="center", va="center")
    if pdf:
        pdf.savefig()
    plt.close(fig)


def advindexing_roll(A: np.ndarray, r: np.ndarray) -> np.ndarray:
    """Shift trace matrix A by r indices.

    Traces and r must be > 1D numpy arrays

    """

    for arg in [A, r]:
        if not isinstance(arg, np.ndarray):
            raise TypeError("Wrong type. Need numpy arrays")

    rows, column_indices = np.ogrid[: A.shape[0], : A.shape[1]]
    r[r < 0] += A.shape[1]
    column_indices = column_indices - r[:, np.newaxis]

    return A[rows, column_indices]


def create_persistent_plot(
    traces: np.ndarray,
    x_range: Optional[Sequence[float]] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
) -> None:
    """Create persistence plot

    Parameters
    ----------

    traces
        2D numpy array (N, M) where N is the number of traces and M the length of the traces
    x_range
        desired range of time axis. If None, it goes from 0 to 2*M in ns
    ax
        a matplotlib axes to plot on. If None, we create a new figure
        and show it at the end. Otherwise this needs to be an array of
        matplotlib axis objects: one axis for each channel present in
        events.

    """
    N = len(traces)
    if not N:
        print("[orange3]WARNING[/] no traces in persistince plot", flush=True)
    mytime = np.linspace(0, 2 * (traces.shape[1] - 1), traces.shape[1])  # in ns

    if x_range is None:
        x_range = [mytime[0], mytime[-1]]
    y_range = [traces.min() * 0.85, traces.max() * 1.15]

    # using datashader-like feature to create persistent plotxlim, ylims
    # height and width of the image
    Nx = traces.shape[1]
    Ny = 400

    data_points_x = mytime[np.newaxis, :].repeat(N, axis=0).flatten()
    data_points_y = traces.flatten()

    img = histogram2d(
        data_points_x, data_points_y, bins=(Nx, Ny), range=(x_range, y_range)
    ).T

    ax_orig = ax
    if ax_orig is None:
        fig, ax = plt.subplots(figsize=(12, 6))
    ax.imshow(
        img + 1e-3,  # add 1e-3 to avoid problems with 0 in LogNorm
        extent=x_range + y_range,
        aspect="auto",
        origin="lower",
        norm=LogNorm(),
        cmap="inferno",
    )
    ax.set_xlabel("Time [ns]")
    ax.set_ylabel("ADC units [a.u]")

    if ax_orig is None:
        plt.show()
        plt.close(fig)

    return img


def find_sums(
    traces: np.ndarray,
    SlowLength: Iterable[int],
    SlowGap: Iterable[int],
    trailsum: Iterable[int],
    gapsum: Iterable[int],
    leadsum: Iterable[int],
) -> Tuple[np.ndarray, np.ndarray]:
    """Find the location of the 3 reported sums in a trace that are used to calculate the energy.

    For this to work, bit 12 of register A (CaptureSums) needs to be set for the channel.

    Parameters
    ----------

    traces
        2D numpy array (N, M) where N is the number of traces and M the length of the traces.
    SlowLength
        List of the length of the Ls for each trace in traces.
    SlowGap
        List of the length of the Gs for each trace in traces.
    trailsum :
        Array/List of N values that contains the values of the first reported sum, which contains L values.
    gapsum :
        Array/List of N values that contains the values of the second reported sum, which contains G values.
    leadsum :
        Array/List of N values that contains the values of the third reported sum, which contains L values.


    Returns
    -------

    sumIDX :
        numpy array of position of the start of the first sumIDX
    Esums :
        numpy array of the three sums

    """
    sumIDX = []
    Esums = []
    for t, Ls, Gs, Tsum, Gsum, Lsum in zip(
        traces, SlowLength, SlowGap, trailsum, gapsum, leadsum
    ):
        for i in range(len(t) - 2 * Ls - Gs):
            s1 = np.sum(t[i : i + Ls])
            s2 = np.sum(t[i + Ls : i + Ls + Gs])
            s3 = np.sum(t[i + Ls + Gs : i + Ls + Gs + Ls])
            # t([s1, s2, s3], [Tsum, Gsum, Lsum], flush=True)
            if [s1, s2, s3] == [Tsum, Gsum, Lsum]:
                sumIDX.append(i)
                Esums.append([s1, s2, s3])
                # print(f'found it at {sumIDX}')
                break
        else:
            print("[red]ERROR[/] Could not find sums!")
            print(
                "[red]ERROR[/] Try increasing trace length and/or trace delay and run again"
            )
            print(
                "Ideal settings for offline computation: trace length > 2*Ls+Gs"
                + " pretrigger length > 3*Ls+Gs (p.69, rev. 01/2019)"
            )
            raise IndexError

    sumIDX = np.array(sumIDX)
    Esums = np.array(Esums)
    return sumIDX, Esums


def MCA(
    data: np.ndarray,
    ax: Optional[matplotlib.axes.Axes] = None,
    rebin: int = 1,
    title: Optional[str] = None,
    label: Optional[str] = None,
    pdf: Optional[PdfPages] = None,
    log: bool = False,
) -> None:
    """Plot MCA spectra

    If pdf is given, it will save the figure inside the pdf (to be used with PdfPages)
    for multi page pdfs.

    Parameters
    ----------

    data :
       Data from ReadHistogramFromModule
    ax
       The axis to plot on, if None, we create a new figure and axis
    rebin
       Rebin the data by summing over this amount of bins
    title
       Add this title to the plot
    pdf
       if present, we save the plot in the pdf
    log
       plot y-axis in log scale

    """
    ax_orig = ax
    if ax is None:
        fig, ax = plt.subplots()
    data = data.reshape(-1, rebin).sum(axis=1)
    ax.plot(data, label=label)
    if log and np.any(data > 0):
        ax.set_yscale("log")
    ax.set_xlabel("Channel")
    ax.set_ylabel("Counts")
    if title:
        ax.set_title(title)
    if pdf:
        pdf.savefig(fig)
    if ax_orig is None:
        plt.close(fig)


def CFD(
    events: Iterable[Event],
    setting: Settings,
    w: float = 0.3125,
    N: Optional[int] = None,
    randomize: bool = False,
    persistent: bool = False,
    title: Optional[str] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    pdf: Optional[PdfPages] = None,
) -> None:
    """Plot CFD and FastFilter several traces

    Parameters
    ----------

    events
        List of events (perhaps already filtered) from read.read_list_mode_data()
        Need to be from a single channel.
    setting
        A read.Settings object
    w
        The `w` used for the CFD calculation
    N
        Number of traces to plot, if `None` all traces will be plotted.
    randomize
        if True and N is smaller than the total amount of traces, randomly pick traces
    persistent
        Use datashader-like plot of traces
    title
        Title to add to plot
    ax
        a matplotlib axes to plot on. If None, we create a new figure
        and show it at the end.
    pdf
       if present, we save the plot in the pdf

    """

    events = filter_events(events, N, randomize)

    channels = {e.channel for e in events}
    if len(channels) != 1:
        print(
            "[red]ERROR[/] Can only plot CFD for events from a single channel at the moment"
        )
        return
    channel = list(channels)[0]

    CFDth = setting.get_by_name("CFDThresh")[channel]
    Lf = int(setting.get_by_name("FastLength")[channel])
    Gf = int(setting.get_by_name("FastGap")[channel])
    FFth = setting.get_by_name("FastThresh")[
        channel
    ]  # TODO: check scaling, perhaps x/L*5

    # parameters needed for CFD calculation
    w = [w]
    B = [5]
    D = [5]
    L = [1]

    traces = [e.trace for e in events]
    for t in traces:
        if t is None:
            print("[red]ERROR[/] Missing trace...skip plotting")
            return
    traces = np.asarray(traces)

    T = np.linspace(0, 2 * (traces.shape[1] - 1), traces.shape[1])  # in ns

    CFD, cfdtime, FF, IDXerr = calculate_CFD_using_FF(
        traces,
        t=T,
        CFD_threshold=CFDth,
        FF_threshold=FFth,
        Lf=Lf,
        Gf=Gf,
        w=w[0],
        B=B[0],
        D=D[0],
        L=L[0],
        Nbkgd=10,
        FF_delay=20,
        CFD_delay=0,
    )

    cfdtime = cfdtime[cfdtime != np.array(None)]

    Tcfd = T[B[0] + D[0] : traces.shape[1] - L[0] + 1]
    Tff = T[0 : FF.shape[1]]

    ax_orig = ax
    if ax_orig is None:
        fig, ax = plt.subplots()
    for t, c, ct, f in zip(traces, CFD, cfdtime, FF):
        ax.plot(T, t, label="Trace", lw=0.2)
        ax.plot(Tcfd, c, label="CFD", lw=0.2)
        ax.plot(Tff, f, label="Fast Filter", lw=0.2)
        ax.plot([T[0], T[-1]], [CFDth, CFDth], label="CFD Threshold", lw=0.2)
        ax.plot([T[0], T[-1]], [FFth, FFth], label="FF Threshold", lw=0.2)
        ax.plot(ct, 0, ".", ms=1, color="k")
    ax.legend(loc="upper right", fancybox=True, shadow=True, prop={"size": 4})

    if title:
        ax.set_title(title)
    if pdf:
        pdf.savefig()
    if ax_orig is None:
        plt.close(fig)


def energy_sums(
    events: Iterable[Event],
    setting: Settings,
    stats: Stats,
    N: Optional[int] = None,
    randomize: bool = False,
    persistent: bool = True,
    title: Optional[str] = None,
    ax: Optional[matplotlib.axes.Axes] = None,
    pdf: Optional[PdfPages] = None,
) -> None:
    """Plots traces aligned to position of the energy sums.

    Plots vertical lines at the position of the energy sums only if all events
    are from a single channel.

    Parameters
    ----------

    events
        List of events (perhaps already filtered) from read.read_list_mode_data()
    setting
        A read.Settings object
    N
        Number of traces to plot, if `None` all traces will be plotted.
    randomize
        if True and N is smaller than the total amount of traces, randomly pick traces
    persistent
        Use datashader-like plot of traces
    title
        Title to add to plot
    ax
        a matplotlib axes to plot on. If None, we create a new figure
        and show it at the end.
    pdf
       if present, we save the plot in the pdf

    """

    events = filter_events(events, N, randomize)

    channels = {e.channel for e in events}

    # create axis if it doesn't exist yet
    ax_orig = ax
    if ax_orig is None:
        fig, ax = plt.subplots()

    # align data
    Ls, Gs, ch, trace, trailsum, gapsum, leadsum = [], [], [], [], [], [], []

    slow_filter_range = setting.get_by_name("SlowFilterRange")
    for e in events:
        if e.CFD_error == 0:
            # convert Ls and Gs into number of data points
            Ls.append(
                int(
                    setting.get_by_name("SlowLength")[e.channel]
                    * (2 ** slow_filter_range)
                    * 10
                    / 2
                )
            )
            Gs.append(
                int(
                    setting.get_by_name("SlowGap")[e.channel]
                    * (2 ** slow_filter_range)
                    * 10
                    / 2
                )
            )
            ch.append(e.channel)
            trace.append(e.trace)
            trailsum.append(e.Esum_trailing)
            gapsum.append(e.Esum_gap)
            leadsum.append(e.Esum_leading)

    trace = np.array(trace)
    try:
        t = np.linspace(0, 2 * (trace.shape[1] - 1), trace.shape[1])  # in ns
    except IndexError:
        return
    IDX = np.arange(0, trace.shape[0], 1)
    if len(trailsum) == 0:
        print("[red]ERROR[/] Raw energy sums must be anabled in the PIXIE-16")
        return

    try:
        sumIDX, Esum = find_sums(trace, Ls, Gs, trailsum, gapsum, leadsum)
    except IndexError:
        for Y in trace:
            ax.plot(t, Y)
        if title:
            ax.set_title(title + " Error in find sums")
        if pdf:
            pdf.savefig()

        if ax_orig is None:
            plt.close(fig)
        return

    minsumIDX = sumIDX.min()
    shiftIDX = sumIDX - minsumIDX
    traceshift = advindexing_roll(trace, -shiftIDX)

    if persistent:
        create_persistent_plot(traceshift[IDX], ax=ax)
    else:
        for Y in traceshift:
            ax.plot(t, Y)

    # draw lines
    if len(channels) == 1:
        positions = [
            t[minsumIDX],
            t[minsumIDX + Ls[0]],
            t[minsumIDX + Ls[0] + Gs[0]],
            t[minsumIDX + Ls[0] + Gs[0] + Ls[0]],
        ]
        ax.vlines(
            positions,
            ymin=trace.min(),
            ymax=trace.max(),
            color="r",
            linestyle="--",
            label="Raw energy sums",
        )
        ch = list(channels)[0]
        live_time = stats["live_time"][ch]
        ax.set_title(
            f"channel: {ch}; number of traces: {len(trace)};"
            f" live time: {live_time:.2f} s"
        )
    else:
        print(
            "[orange3]WARNING[/] plot.energy_sums: more than one channel given, not drawing positions of sums"
        )
    ax.legend()
    ax.set_ylim(bottom=0)
    ax.legend()

    if title:
        ax.set_title(title)
    if pdf:
        pdf.savefig()

    if ax_orig is None:
        plt.close(fig)



"""
Usage:
   create_variable_list <filename>

Created variables.py which is used by the module to figure out where
in memory different settings are located.

"""

import docopt

commands = docopt.docopt(__doc__)

filename = commands["<filename>"]

with open(filename) as f:
    lines = f.readlines()

values = {}
last_idx = None
for l in lines:
    a, b = l.strip().split()
    # need the ,0 to be convert from hex to int
    a = int(a, 0) - 0x4A000
    values[b] = a
    if last_idx:
        start = values[last_idx]
        values[last_idx] = [start, a - start]
    last_idx = b
if last_idx:
    start = values[last_idx]
    values[last_idx] = [start, 16]

with open("variables.py", "w") as f:
    f.write(
        '# automatically created by running "create_variable_list.py Pixie16DSP_revfgeneral_14b500m_r35207.var"\n'
    )
    f.write("# do not modify manually unless you know what you are doing ;)\n\n")

    f.write("settings = {\n")
    for k, v in values.items():
        f.write(f'          "{k}": [{v[0]}, {v[1]}],\n')
    f.write("        }\n")



"""Task and Pipeline class to run data acquisition and data handling.

This file provides two base classes: Task and Pipeline.

They can be used to create a data acquisition pipeline to take data
with the pixie16, convert the binary data to event data and do some
calculation on them.

To use the classes, create custom Task that inherit the base
class. Overwrite the init function if needed and the do_work
function to process the current work load. The cleanup function can be
overwritten to close files and is called once the tasks is finished or
if it gets a keyboard interrupt.

Tasks can then be chained together automatically using the Pipeline
class. This class can also be used to run the pipeline, print a nice
status bar. An example use of 3 different tasks is:

    A = TaskA()
    B = TaskB()
    C = TaskC()
    tasks = [A, B, C]

    pipeline = Pipeline(tasks)

    pipeline.start()  # starts all tasks

    pipeline.wait_with_progress(total=14.0)  # reports progress with a bar

    pipeline.join()  # waits for all tasks to finish
    pipeline.close()  # releases recsources

"""

from contextlib import redirect_stdout
import io
import multiprocessing
import queue
import sys
import time
import traceback
from typing import List, Union

from more_itertools import pairwise
from rich.progress import Progress, BarColumn, TimeRemainingColumn, TimeElapsedColumn
from rich import print


class Task(multiprocessing.Process):
    """Custom Process that can be stopped and joined into a pipeline.

    Queues are used to get data or send data to the next Tasks. We
    provide functions that should be overwritten when customizing
    a Task:

    do_work(value): gets called continously in a while loop without an input queue
    or gets called whenever a value arrives in the input queue.
    If the tasks is done, do_work should set self.done to True.

    cleanup(): Will be called at the end of a tasks, option to close any files, etc.

    If you start a data acquisition in the pixie, make sure that you call
    `init_and_boot` inside the do_work function. Since the, for example, the `__init__`
    of a Task will still be executed in the main thread.

    For example of Tasks, see `task.py`

    Since Tasks are meant to be daisychained together, each has an `input_queue` and
    an `output_queue`. Furthermore, there can be a `status_queue` that is used to talk
    to the main thread and print information to the screen.

    Any `print` output that is generated in the do_work function will be captured and
    send as a message to the main pipeline (using the `status_queue`) thread where it
    will be printed using rich.

    """

    def __init__(self, event=None):
        super().__init__()
        self.stop_event = event or multiprocessing.Event()
        self.input_queue = None
        self.output_queue = None
        self.status_queue = None
        self.start_time = None
        self.stop_time = None
        self.messages = []  # messages to be send to the main pipeline
        self.name = "general task"
        self.done = False
        self.last_update = time.time()

    def cleanup(self):
        """Can be overwritten to close any open files, etc"""
        pass

    def send_status(self, value_dict):
        """Send a status update to the main pipeline.

        Status updates are only send at least 0.5 seconds apart so
        that they don't spam the status queue.

        Since sending of the `value_dict` is not garantueed, do not rely on
        this feature and only send information for information purposes.

        Currently, there are two keys in the `value_dict` that have special meaning:
        - runtime: should be the elappsed time during data acquisition
        - message: Any message that should be printed to the screen (we use this, so
                   that the progress bar works well)

        """
        current_time = time.time()
        if self.status_queue and (current_time - self.last_update > 0.5):
            self.status_queue.put(value_dict)
            self.last_update = current_time
            for m in self.messages:
                self.status_queue.put({"message": m})
            self.messages = []

    def do_work(self, value):
        """Needs to be overwritten to do the work.

        value will be the current element in the queue.
        """
        raise NotImplementedError

    def stop(self):
        """Set signal that will stop all workers."""
        self.stop_event.set()

    def join(self, *args, **kwargs):
        """Stop all workers and wait until they are done."""
        self.stop()
        super().join(*args, **kwargs)

    def handle_print(self, stringIO: io.StringIO):
        """Handle print statements that got captured"""
        msg = stringIO.getvalue()
        if msg:
            msg = msg.split("\n")
            self.messages += msg
        self.send_status({})

    def run(self):
        try:
            while not self.stop_event.is_set() and not self.done:
                # get next value from queue or if no queue set it to None
                if self.input_queue is None:
                    value = None
                else:
                    try:
                        value = self.input_queue.get(timeout=0.2)
                        if value is None:
                            self.done = True
                            continue
                    except queue.Empty:
                        continue
                # save the time we do the first work unit
                if self.start_time is None:
                    self.start_time = time.time()
                # do the work and capture any print statements
                with redirect_stdout(io.StringIO()) as f:
                    out = self.do_work(value)
                    self.handle_print(f)

                # send result to next stage if available
                if self.output_queue and out is not None:
                    self.output_queue.put(out)
        except KeyboardInterrupt:
            self.stop_event.set()
        except Exception as e:
            print(f"[DEBUG] ----- exception in {self.name} ----")
            print(e)
            print(sys.exc_info())
            traceback.print_exc()
            sys.stdout.flush()
            sys.stderr.flush()
            self.stop_event.set()
        finally:
            with redirect_stdout(io.StringIO()) as f:
                self.cleanup()
                self.handle_print(f)

        # if we are done, indicate that the queue should end
        if self.output_queue:
            self.output_queue.put(None)
        self.stop_time = time.time()

        # print(f"Finished Tasks {self.name}")


class Pipeline:
    """Manage a linear chain of tasks.

    Chain the input and outputs together, some shortcuts to
    start/stop/join all the Tasks and to test if any Task is still
    running.

    """

    def __init__(self, tasks: List[Task], name: str = "Run", verbose: bool = False):
        self._tasks = tasks
        self.verbose = verbose
        self.status = {}
        self.messages = []
        self.i = 0
        self.start_time = 0
        self.runtime = None
        self.queues = {}
        self.name = name

        # we'll use a single event for all tasks
        self.stop_event = multiprocessing.Event()
        # set up a Queue that can report back status to the main task
        self.status_queue = multiprocessing.Queue()

        for t in self._tasks:
            t.stop_event = self.stop_event
            t.status_queue = self.status_queue

        if len(tasks) > 1:
            for A, B in pairwise(tasks):
                self.link_tasks(A, B)

    def link_tasks(self, A, B):
        """Link two tasks in a pipeline using a queue."""
        q = multiprocessing.Queue()
        A.output_queue = q
        B.input_queue = q
        self.queues[A.name] = q
        if self.verbose:
            print(
                f"[Pipeline] setting up pipeline link: {A.name} -> {B.name}", flush=True
            )

    def start(self):
        """Start all tasks."""
        self.start_time = time.time()
        for t in self._tasks:
            t.start()

    def join(self):
        """Join all tasks, i.e. wait until they are all done."""
        for t in self._tasks:
            t.join()

    def close(self):
        """Release all resources, needed for pixie16.control."""
        for t in self._tasks:
            t.close()
        for q in self.queues.values():
            q.close()

    def stop(self):
        """Stop all workers in the pipeline."""
        self.stop_event.set()

    def is_alive(self):
        """Check if any task is still running."""
        for t in self._tasks:
            if t.is_alive():
                return True
        return False

    def execute(self, progress_bar_time: Union[float, int] = 0):
        """Runs the whole pipeline.

        Parameters
        ----------
        progress_bar_time
           if > 0 run with a progress bar with this number as the total time it should
           take. Otherwise, run without progressbar.

        """
        self.start()
        if progress_bar_time:
            self.wait_with_progress(total=progress_bar_time)
        else:
            self.wait()
        self.join()
        self.close()

    def update_status(self):
        """Look at the status queue and update the pipeline status.

        We also parse out the runtime of the pixie16 if present.
        """
        try:
            while True:
                value = self.status_queue.get(block=False)
                if "runtime" in value:
                    self.runtime = value.pop("runtime")
                if "message" in value:
                    self.messages.append(value.pop("message"))

                self.status.update(value)
        except queue.Empty:
            pass

    def wait_with_progress(self, total):
        """Create a progress bar while the pipeline is running and show the status updates.

        The run can be interruptted using Ctrl-C.
        """
        try:
            with Progress(
                f"[green]{self.name}",
                BarColumn(
                    style="grey35 on black",
                    complete_style="green on black",
                    finished_style="red on black",
                    pulse_style="white on purple",
                ),
                TimeElapsedColumn(),
                TimeRemainingColumn(),
                "{task.description}",
                refresh_per_second=4,
            ) as progress:
                task = progress.add_task("[green] Run", total=total)
                while self.is_alive():
                    time.sleep(0.1)
                    self.update_status()
                    for m in self.messages:
                        progress.console.print(m)
                    self.messages = []
                    text = ",".join(f"{k}: {v}" for k, v in self.status.items())
                    if self.runtime is None:
                        text = "Setting up process pipeline " + text
                        runtime = round(time.time() - self.start_time, 2)
                    else:
                        runtime = self.runtime
                    progress.update(task, completed=runtime, description=text)
                # finish progress bar
                progress.update(task, completed=total, description=text)
        except KeyboardInterrupt:
            print("[yellow]INFO[/] got keyboard interrupt... closing")
            self.stop()

    def wait(self):
        """Wait for the pipeline to finish.

        The run can be interruptted using Ctrl-C.
        """
        try:
            while self.is_alive():
                time.sleep(0.1)
        except KeyboardInterrupt:
            print("[yellow]INFO[/]] got keyboard interrupt... closing")
            self.stop()



from PyQt5.QtCore import QAbstractTableModel, Qt
from PyQt5.QtWidgets import QCommonStyle, QStyle

import pixie16


class BinaryDataModel(QAbstractTableModel):
    """Maps to a list of pixie16.read.Event data.

    E.g. data from a list mode run of the pixie16.

    """

    def __init__(self, data):
        super().__init__()
        self._data = data
        self.header = pixie16.read.Event._fields
        self.icon_true = QCommonStyle().standardIcon(QStyle.SP_DialogApplyButton)
        self.icon_false = QCommonStyle().standardIcon(QStyle.SP_DialogCancelButton)

    def icon_return(self, value):
        if value:
            return self.icon_true
        return self.icon_false

    def data(self, index, role):
        """Return the data.

        For some True/False data, we display an icon instead. To do
        this, we return no data in DisplayRole, but an icon for
        DecorationRole.

        """
        column = index.column()
        row = index.row()

        if role == Qt.DisplayRole:
            if self.header[column] in [
                "trace",
                "CFD_error",
                "pileup",
                "trace_flag",
            ]:
                return None
            return str(self._data[row][column])
        if role == Qt.DecorationRole:
            if self.header[index.column()] == "trace":
                return self.icon_return(len(self._data[row][column]))
            if self.header[index.column()] in ["CFD_error", "pileup", "trace_flag"]:
                return self.icon_return(not self._data[row][column])

    def rowCount(self, index):
        return len(self._data)

    def columnCount(self, index):
        return len(self.header)

    def headerData(self, section, orientation, role):
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return self.header[section].title()



def compare_module_setting(
    A, B, quiet=False, writeable_only=True, units=True, channels=None
):
    """Compare the settings of two modules

    Print the difference.

    If quiet is True, then just return True/False

    The functions assumes that both settings have the same structure.
    """

    if type(A) != type(B):
        print(f"compare_module_setting: both items need to have the same type")
        return None

    if type(A) != Settings:
        print(f"compare_module_setting: both items need to be Setting classes")
        return None

    if not quiet:
        print("Comparing settings:")
        print(f"  a: {A.filename}")
        print(f"  b: {B.filename}")
        print("")

        if writeable_only:
            keys = [k for k in settings.keys() if settings[k][0] < 832]
        else:
            keys = list(settings.keys())

        if units:
            keys = keys + Settings.derived_keys
            keys = [k for k in keys if k not in Settings.derived_from_keys]

        for k in keys:
            if units:
                valueA = A.units[k]
                valueB = B.units[k]
            else:
                valueA = A[k]
                valueB = B[k]
            if isinstance(valueA, (int, str, float)):
                if valueA != valueB:
                    print(f"{k} differs:\n     a: {valueA}\n     b: {valueB}")
            elif isinstance(valueA, (list, tuple)):
                first = True
                for i, (a, b) in enumerate(zip(valueA, valueB)):
                    if channels:
                        if i not in channels:
                            continue
                    if a != b:
                        if first:
                            print(f"{k} differs:")
                            first = False
                        # if these are bit values, only show bits that are different
                        if isinstance(a, str) and "-" in a:
                            aa = a.split("-")
                            bb = b.split("-")
                            for x, y in zip(aa, bb):
                                if x != y:
                                    print(f"  ch:{i}\n     a: {x}\n     b: {y}")
                        else:
                            print(f"  ch:{i}\n     a: {a}\n     b: {b}")
            else:
                print("This shouldn't happen. type value A", type(valueA))

    # convert to a sorted json string to make deep comparison possible
    if writeable_only:
        return json.dumps(A.rawdata[:832], sort_keys=True) == json.dumps(
            B.rawdata[:832], sort_keys=True
        )
    else:
        return json.dumps(A.rawdata, sort_keys=True) == json.dumps(
            B.rawdata, sort_keys=True
        )



"""Classes for each field inside a setings file.

These contain helper functions to, for example, combine the HI and LO values of timestamps.
"""

import abc
from typing import Tuple


class NotImplementedAttr:
    """An 'abstract attribute' class that allows for initialization for an owner instance but not access to this
    attribute"""

    def __init__(self, err_msg):
        self.err_msg = err_msg

    def __get__(self, instance, owner):
        raise NotImplementedError(self.err_msg)


class SettingsFieldBase(abc.ABC):
    """A descriptor base class that will get derived settings values according to some logic. Said logic should be
    defined in a subclass's implementation of derive_field_value"""

    name = NotImplementedAttr(
        "A SettingsFieldBase instance can only be defined as an attribute in a subclass of Entity"
    )

    @property
    def storage_names(self) -> Tuple[str]:
        """Name of the raw settings file field(s) used to store the value of this derived field. Default is set to the
        name of this field."""
        return (self.name,)

    def get_storage_values(self, instance):
        """Utility function to grab the values of the storage fields zipped together."""
        storage_values = []
        for sn in self.storage_names:
            # If the storage name is the name of this field, grab the value with dict syntax to avoid infinite recursion
            if sn == self.name:
                sv = instance.__dict__[sn]
            # If not, use getattr so that we can also grab derived fields (whose values are not directly stored in
            # __dict__)
            else:
                sv = getattr(instance, sn)
            try:
                if isinstance(sv, str):
                    raise NotImplementedError(
                        "Did not expect any settings to be saved as string"
                    )
                sv = list(sv)
            except TypeError:
                sv = [sv]
            storage_values.append(sv)
        return zip(*storage_values)

    @abc.abstractmethod
    def derive_field_value(self, instance, *storage_values):
        """An abstract method that should implement the logic to derive the value of this field from the raw settings
        file values."""

    def __get__(self, instance, owner):
        """Get the value of this field. All the logic for how to combine raw settings values is left to
        derive_field_values"""
        storage_values = self.get_storage_values(instance)
        res = tuple(self.derive_field_value(instance, *v) for v in storage_values)
        if len(res) == 1:
            return res[0]
        else:
            return res

    def __set__(self, instance, value):
        """By default SettingsFields are not settable."""
        raise AttributeError("The set method is not defined for this attribute")


class EntityMeta(type):
    """Metaclass for settings class. Assigns the name of the variable that holds the instances of SettingsFieldBase,
    call one such instance s, to s.name. E.g. if apple = SettingFieldSubclass() were defined in an EntityMeta class then
    apple.name would be 'apple'. Adapted from the Fluent Python book"""

    def __init__(cls, name, bases, attr_dict):
        super().__init__(name, bases, attr_dict)
        for key, attr in attr_dict.items():
            if isinstance(attr, SettingsFieldBase):
                attr.name = key


class Entity(metaclass=EntityMeta):
    """Entity with settings fields"""


# ----------------------------------------------------------------------------------------------------------------------


def combine_hi_lo_registers(hi, lo):
    return hi * 2 ** 32 + lo


class ABField(SettingsFieldBase):
    """A SettingsField to derives fields separated into hi lo 32 bit register subfields. Referred to as 'AB' classes
    because such fields are named 'FastPeaksA' and 'FastPeaksB' for example."""

    @property
    def storage_names(self) -> Tuple[str]:
        return self.name + "A", self.name + "B"

    def derive_field_value(self, instance, *storage_values):
        A, B = storage_values
        return combine_hi_lo_registers(A, B)


def int_to_bits(i, bits_length):
    """Converts int values to binary (in the form of a tuple of 0's and 1's)"""
    bit_fmt = f"{{:0{bits_length}b}}"
    return tuple(int(j) for j in bit_fmt.format(i))


BIT_FIELD_SIZE_TBL = {
    "MultiplicityMaskL": 32,
    "MultiplicityMaskH": 32,
    "FastTrigBackplaneEna": 32,
    "TrigConfig": 32,
    "ChanCSRa": 22,
    "ChanCSRb": 22,
    "ModCSRA": 14,
    "ModCSRB": 14,
}


class BinaryField(SettingsFieldBase):
    """A SettingsField to derive values that are meant to be interpreted as binary digits but are stored as ints"""

    def derive_field_value(self, instance, *storage_values):
        (int_value,) = storage_values
        bits_length = BIT_FIELD_SIZE_TBL[self.name]
        return int_to_bits(int_value, bits_length)


def calc_trace_delay(paf_length, trigger_delay):
    return paf_length - trigger_delay


class TraceDelayField(SettingsFieldBase):
    """A SettingsField to derive trace delay"""

    storage_names = ("PAFlength", "TriggerDelay")

    def derive_field_value(self, instance, *storage_values):
        paf_length, trigger_delay = storage_values
        return calc_trace_delay(paf_length, trigger_delay)


def calc_offset_voltage(offset_dac):
    return 1.5 * ((32768 - offset_dac) / 32768)


class OffsetVoltageField(SettingsFieldBase):
    """A SettingsField to derive offset voltage"""

    storage_names = ("OffsetDAC",)

    def derive_field_value(self, instance, *storage_values):
        (offset_dac,) = storage_values
        return calc_offset_voltage(offset_dac)



"""
Usage:
  pixie16-parameter-scan <settingfile> (<parameter> <module> <channel> <start> <stop> <step>)...
  pixie16-parameter-scan [options]

Loads the settingfile and then scans the pixie16 parameter.

start, stop and step are in ns and will be converted to μs internally.

Live 1D and 2D plots are being displayed and the images and settingfiles saved.

Options:
  --list-parameters   print a list of all possible parameters

"""

import docopt
import sys


def main():
    commands = docopt.docopt(__doc__)
    # print(commands)

    # only import now, since some of the imports take some time
    # this way the command line feels a bit faster
    from pixie16 import scan as pxscan
    from pixie16 import control as pxcontrol

    from pathlib import Path
    import datetime

    if commands["--list-parameters"]:
        print("name                 unit          DPS name")
        print("-------------------------------------------")
        for n, u, dps in pxscan.PARAMETERS:
            uu = f"[{u}]"
            print(f"{n:20} {uu:13} {dps:20}")
        sys.exit()

    today = f"{datetime.datetime.now():%Y-%m-%d}"
    datadir = Path(".")

    # Initialize System, load config file
    pxcontrol.init_and_boot(modules=[2])
    pxcontrol.load_dsp_parameters_from_file(Path(commands["<settingfile>"]))

    # Run info
    dwelltime = 1  # in seconds

    # Write Run Time
    for i, m in enumerate(pxcontrol.modules):
        tmp = pxcontrol.set_run_time(dwelltime, i)
        print(f"Run Time: {tmp}")

    N = len(commands["<channel>"])

    run = 0
    for f in datadir.glob(f"{today}-scan-*.png"):
        parts = f.name.split("-")
        current_run = int(parts[-1][:-4])
        if current_run >= run:
            run = current_run + 1
    print(f"Run: {run}")

    plot_data = datadir / f"{today}-scan-{run:03d}.png"

    if N == 1:
        parameter = pxscan.ScanParameter(
            name=commands["<parameter>"][0],
            module=int(commands["<module>"][0]),
            channel=int(commands["<channel>"][0]),
            start=float(commands["<start>"][0]),
            stop=float(commands["<stop>"][0]),
            step=float(commands["<step>"][0]),
        )

        title = f'1d-Scan using {commands["<settingfile>"]}'
        pxscan.parameter_scan_1d(parameter, title=title, savefile=plot_data)

    elif N == 2:
        parameterA = pxscan.ScanParameter(
            name=commands["<parameter>"][0],
            module=int(commands["<module>"][0]),
            channel=int(commands["<channel>"][0]),
            start=float(commands["<start>"][0]),
            stop=float(commands["<stop>"][0]),
            step=float(commands["<step>"][0]),
        )
        parameterB = pxscan.ScanParameter(
            name=commands["<parameter>"][1],
            module=int(commands["<module>"][1]),
            channel=int(commands["<channel>"][1]),
            start=float(commands["<start>"][1]),
            stop=float(commands["<stop>"][1]),
            step=float(commands["<step>"][1]),
        )

        title = f'2d-Scan using {commands["<settingfile>"]}'

        pxscan.parameter_scan_2d(
            parameterA, parameterB, title=title, savefile=plot_data
        )
    else:
        print("Only 1d and 2d scans are implemented")

    newfile = datadir / f"{today}-scan-{run:03d}-setting.json"
    pxcontrol.save_dsp_parameters_as_json(newfile)
    newfile = datadir / f"{today}-scan-{run:03d}-stats.json"
    pxcontrol.save_stats_to_file(newfile)

    pxcontrol.end_run()
    pxcontrol.exit_system()
